%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DFBAlab: Dynamic Flux Balance Analysis laboratory                       %
% Process Systems Engineering Laboratory, Cambridge, MA, USA              %
% July 2014                                                               %
% Written by Jose A. Gomez and Kai Hffner                                %
%                                                                         % 
% This code can only be used for academic purposes. When using this code  %
% please cite:                                                            %
%                                                                         %
% Gomez, J.A., Hffner, K. and Barton, P. I.                              %
% DFBAlab: A fast and reliable MATLAB code for Dynamic Flux Balance       %
% Analysis. Submitted.                                                    %
%                                                                         %
% COPYRIGHT (C) 2014 MASSACHUSETTS INSTITUTE OF TECHNOLOGY                %
%                                                                         %
% Read the LICENSE.txt file for more details.                             %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Clear workspace

clear all
tic
% Add paths to LP solver and DFBAlab
  
% addpath(genpath('C:\Program Files\Gurobi'))
% addpath(genpath('C:\Users\Mike\Documents\MATLAB\DFBAlab_0'))

addpath(genpath('C:\Program Files (x86)\Gurobi605'))
addpath(genpath('N:\poonam\OneDrive\D drive\Matlab\DFBAlab_0\DFBAlab_0'))


% Specify number of discretization points

nspecies = 2;
npoints = 5;
nmodel = nspecies*npoints;
INFO.nmodel = nmodel;

% Load models. These should be .mat files generated by the COBRA toolbox. 
% When generating these files using the COBRA toolbox, a big number is used
% as infinity. This number should be fed to the DB vector (Default bound).
    
load model.mat
pa.lb(358) = 5;
for i = 1:npoints
    model{i} = pa;
    DB(i) = 1000;
    model{npoints+i} = sa;
    DB(npoints+i) = 1000;
end
INFO.DB = DB;

%% exID array
% You can either search the reaction names by name or provide them directly
% in the exID array.
% RxnNames = {'EX_glc(e)', 'EX_ac(e)', 'biomass'};
% for i = 1:length(RxnNames)
%    [a,exID(i)] = ismember(RxnNames(i),model.rxns);
% end

for i = 1:npoints
    exID{i}=[88 135 177 95 197 154 1112];
    exID{npoints+i}=[1 604 601 564 621 559];
end
INFO.exID = exID;

% Lower bounds and upper bounds for these reactions should be provided in
% the RHS code. 

% This codes solves the LPs in standard form. Bounds on exchange fluxes in 
% the exID array can be modified directly on the first 2*n rows where n is 
% the number of exchange fluxes. Order will be lower bound, upper bound, 
% lower bound, upper bound in the same order as exID. 

% NOTE: All bounds on fluxes in the exID arrays are relaxed to -Inf and 
% + Inf. These bounds need to be updated if needed in the RHS file.

%% Cost vectors
% Usually the first cost vector will be biomass maximization, but it can
% be any other objective. The CPLEX objects will minimize by default. 
% Report only nonzero elements. 
% The structure should be:
% C{model} = struct
% Element C{k}(i) of C, is the cost structure i for model k. 
% C{k}(i).sense = +1 for minimize, or -1 for maximize.
% C{k}(i).rxns = array containing the reactions in this objective. 
% C{k}(i).wts = array containing coefficients for reactions reported in 
% rxns. Both arrays should have the same length. 
% Example, if:
% C{k}(i).rxns = [144, 832, 931];
% C{k}(i).wts = [3, 1, -1];
% Then the cost vector for this LP will be:
% Cost{k}(i) = 3*v_144 + v_832 - v_931 (fluxes for model k). 
% This cost vector will be either maximized or minimized depending on the
% value of C{k}(i).sense.

% In SBML files, usually production fluxes are positive and uptake fluxes
% are negative. Keep in mind that maximizing a negative flux implies 
% minimizing its absolute value.
% Different models can have different number of objectives. 

minim = 1;
maxim = -1;

for i=1:npoints
    
    % Maximize PA growth
    C{i}(1).sense = maxim;
    C{i}(1).rxns = [88];
    C{i}(1).wts = [1];
    
    % Minimize PA acetate synthesis 
    C{i}(2).sense = minim;
    C{i}(2).rxns = [95];
    C{i}(2).wts = [1];
    
    % Minimize PA succinate synthesis
    C{i}(3).sense = minim;
    C{i}(3).rxns = [197];
    C{i}(3).wts = [1];
    
    % Maximize PA glucose uptake
    C{i}(4).sense = minim;
    C{i}(4).rxns = [135];
    C{i}(4).wts = [1];
    
    % Maximize PA oxygen uptake
    C{i}(5).sense = minim;
    C{i}(5).rxns = [177];
    C{i}(5).wts = [1];
    
    % Maximize PA lactate uptake
    C{i}(6).sense = minim;
    C{i}(6).rxns = [154];
    C{i}(6).wts = [1];
   
    % Minimize PA pyocyanin synthesis
    C{i}(7).sense = minim;
    C{i}(7).rxns = [1112];
    C{i}(7).wts = [1];
    
    % Maximize SA growth
    C{npoints+i}(1).sense = maxim;
    C{npoints+i}(1).rxns = [1];
    C{npoints+i}(1).wts = [1];
  
    % Minimize SA acetate synthesis
    C{npoints+i}(2).sense = minim;
    C{npoints+i}(2).rxns = [564];
    C{npoints+i}(2).wts = [1];
    
    % Maximize SA lactate synthesis
    C{npoints+i}(3).sense = maxim;
    C{npoints+i}(3).rxns = [559];
    C{npoints+i}(3).wts = [1]; 
    
    % Maximize SA glucose uptake
    C{npoints+i}(4).sense = minim;
    C{npoints+i}(4).rxns = [604];
    C{npoints+i}(4).wts = [1];
    
    % Maximize SA oxygen uptake 
    C{npoints+i}(5).sense = minim;
    C{npoints+i}(5).rxns = [601];
    C{npoints+i}(5).wts = [1];
    
    % Maximize SA succinate uptake 
    C{npoints+i}(6).sense = minim;
    C{npoints+i}(6).rxns = [621];
    C{npoints+i}(6).wts = [1];
    
end
INFO.C = C;    

% Set number of discretization points and state variables

N = npoints-1; 
ns = 8;
INFO.N = N;
INFO.ns = ns;

% Time of simulation

tspan = [0,1000];

% Biofilm conditions and parameters

L = 80;                   % Biofilm thickness, micron
x0 = 1;                   % PA initial biomass concentration, g/L
z0 = 1;                   % SA initial biomass concentration, g/L
glcb = 7.5;               % Exudate glucose concentration, mmol/L
o2L = 0.21;               % Air oxygen concentration, mmol/L
aceb = 0;                 % Exudate acetate concentration, mmol/L
succb = 0;                % Exudate succinate concentration, mmol/L
lacb = 1;                 % Exudate lactate concentration, mmol/L
pyob = 0;                 % Initial inhibitor concentration, mmol/L
glcdiff = 9.4e-6;         % Aqueous glucose diffusion coefficient, cm^2/s
o2diff =  26.8e-6;        % Aqueous oxygen diffusion coefficient, cm^2/s
acediff = 16.2e-6;        % Aqueous acetate diffusion coefficient, cm^2/s
succdiff = 12.6e-6;       % Aqueous succinate diffusion coefficient, cm^2/s
lacdiff = 12.1e-6;        % Aqueous lactate diffusion coefficient, cm^2/s
pyodiff = 7.2e-6;         % Aqueous inhibitor diffusion coefficient, cm^2/s
kxt = 5e-8;               % PA aerotaxis coefficient, units?
glcmt = 2e-4;             % Glucose mass transfer coefficient, cm/s
o2mt = 2e-2;              % Oxygen mass transfer coefficient, cm/s
acemt = 8e-4;             % Acetate mass transfer coefficient, cm/s
succmt = 8e-4;            % Succinate mass transfer coefficient, cm/s
lacmt = 8e-4;             % Lactate mass transfer coefficient, cm/s
pyomt = 2e-4;             % Inhibitor mass transfer coefficient, cm/s
celldiff = 0;%1e-12; %0.05;  % PA inhibitor synthesis rate, 1/h
kpyod = 0.4;              % SA death rate due to pyocyanin, 1/h units??
kpadm = 0.01;             % PA death rate constant, 1/h
ksadm = 0.01;             % SA death rate constant, 1/h
xmax = 200;               % Combined maximum biomass concentration, g/L
zs = L/N;                 % Spatial step size, micron
td = 1;                   % Characteristic time, hr

% Set parameter vector

condit = [L,glcdiff,o2diff,acediff,succdiff,lacdiff,pyodiff,kxt,glcmt,o2mt,acemt,...
    succmt,lacmt,pyomt,xmax,N,ns,zs,glcb,o2L,aceb,succb,lacb,pyob,celldiff,kpyod,kpadm,ksadm,td];
INFO.condit=condit;

% Uptake scenario and parameters

uptake = [1       % PA glucose uptake
          1       % PA oxygen uptake
          0       % PA acetate uptake
          0       % PA succinate uptake
          0       % PA lactate uptake
          1       % SA glucose uptake
          1       % SA oxygen uptake
          0       % SA acetate uptake
          0       % SA succinate uptake
          0       % SA lactate uptake
         ];

param = [
     10              % PA maximum glucose uptake rate, mmol/g/h 
     0.5             % PA glucose uptake saturation constant, mmol/L 
     20              % PA maximum oxygen uptake rate, mmol/g/h  
     0.003           % PA oxygen uptake saturation constant, mmol/L 
     10              % PA maximum acetate uptake rate, mmol/g/h 
     0.5             % PA acetate uptake saturation constant, mmol/L
     10              % PA maximum succinate uptake rate, mmol/g/h
     0.5             % PA succinate uptake saturation constant, mmol/L
     10              % PA maximum lactate uptake rate, mmol/g/h
     0.5             % PA lactate saturation constant, mmol/L
     10              % SA maximum glucose uptake rate, mmol/g/h 
     0.5             % SA glucose uptake saturation constant, mmol/L 
     20              % SA maximum oxygen uptake rate, mmol/g/h  
     0.003           % SA oxygen uptake saturation constant, mmol/L 
     10              % SA maximum acetate uptake rate, mmol/g/h 
     0.5             % SA acetate uptake saturation constant, mmol/L
     10              % SA maximum succinate uptake rate, mmol/g/h
     0.5             % SA succinate uptake saturation constant, mmol/L
     10              % SA maximum lactate uptake rate, mmol/g/h
     0.5             % SA lactate saturation constant, mmol/L
    ];

param(1:2:20) = uptake(1:10).*param(1:2:20);
INFO.param=param;


% Specify initial conditions for each discretization point
% Initial concentrations of PA biomass (g/L), SA biomass (g/L), glucose (mmol/L),
% oxygen (mmol/L), acetate (mmol/L), succinate (mmol/L), lactate (mmol/L), inhibitor (mmol/L) 

yo = [];
for i = 1:N+1
    yz = [x0,z0,glcb,o2L,aceb,succb,lacb,pyob];
    yo = [yo yz];
end

% yo = pa_sa_ic_N50_L80_atp5;

% Add initial condition for penalty variable

yo = [yo zeros(1,nmodel)];
Y0 = yo;

% CPLEX Objects construction parameters

INFO.LPsolver = 1; % CPLEX = 0, Gurobi = 1.
                   % CPLEX works equally fine with both methods.
                   % Gurobi seems to work better with Method = 1, and 
                   % Mosek with Method = 0.
                   
INFO.tol = 1E-9; % Feasibility, optimality and convergence tolerance for Cplex (tol>=1E-9). 
                 % It is recommended it is at least 2 orders of magnitude
                 % tighter than the integrator tolerance. 
                 % If problems with infeasibility messages, tighten this
                 % tolerance.
                 
INFO.tolPh1 = INFO.tol; % Tolerance to determine if a solution to phaseI equals zero.
                        % It is recommended to be the same as INFO.tol. 
                        
INFO.tolevt = 1E-6; % Tolerance for event detection. Has to be greater
                          % than INFO.tol.

% You can modify the integration tolerances here.
% If some of the flows become negative after running the simulation once
% you can add the 'Nonnegative' option.
% Include function to compute non-zero Jacobian entries.

NN = 1:length(Y0);
options = odeset('AbsTol',5E-7,'RelTol',5E-7,'NonNegative',NN,...
    'Events',@evts,'JPattern',jac_pa_biofilm(N,ns));

% INFO: You can use the INFO struct to pass parameters. Don't use any of 
% the names already declared or: INFO.t (carries time information), 
% INFO.ncost, INFO.lexID, INFO.LlexID, INFO.lbct, INFO.ubct, INFO.sense,
% INFO.b, INFO.pair. 

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%+%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
[model,INFO] = ModelSetupM(model,Y0,INFO);
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
INFO.INFtimes = zeros(1,nmodel);
if INFO.LPsolver == 0
    [INFO] = LexicographicOpt(model,INFO);
elseif INFO.LPsolver == 1
    [INFO] = LexicographicOptG(model,INFO);
else
    display('Solver not currently supported.');
end

tint = 0;
TF = [];
YF = [];
flux = [];
while tint<tspan(2)
    
% Look at MATLAB documentation if you want to change solver.
% ode15s is more or less accurate for stiff problems.
    INFO.INFtimes;
    [T,Y] = ode15s(@odes_pa_sa_biofilm,tspan,Y0,options,INFO);
    for i=1:length(T)
        [dy,fluxy] = odes_pa_sa_biofilm(T(i),Y(i,:),INFO);
        flux = [flux;fluxy];
    end
    TF = [TF;T];
    YF = [YF;Y];
    tint = T(end);
    tspan = [tint,tspan(2)];
    Y0 = Y(end,:);

    if tint == tspan(2)
        break;
    end
    
% Update b vector 
[INFO] = bupdate(tint,Y0,INFO);
%Determine model with basis change
    value = evts(tint,Y0,INFO);
    ind = find(value(1:end-nmodel)<=0);
    fprintf('Basis change at time %d. ',tint);
    k = 0;
    ct = 0;
    
% Detect model infesibilities
    INFvector =  value(end-nmodel+1:end);
    for i=1:length(INFvector)
        if INFvector(i)>0 && INFO.INFtimes(i) == 0
            INFO.INFtimes(i) = tint;
            fprintf('Model %i is now dying at time %d. \n',i,tint);
        end
    end
    
    while ~isempty(ind)
        k = k + 1;
        ct = ct + size(model{k}.A,1);
        ind2 = find(ind<=ct);
        if ~isempty(ind2)
           INFO.flagbasis = k; 
           fprintf('Model %i. \n',k);
           % Perform lexicographic optimization
           if INFO.LPsolver == 0
               [INFO] = LexicographicOpt(model,INFO);
           elseif INFO.LPsolver == 1
               [INFO] = LexicographicOptG(model,INFO);
           else
               display('Solver not currently supported.');
           end
           ind(ind2)=[];
        end
    end
end

T = TF;
Y = YF;

toc
% Print variables at the tissue-biofilm and biofilm-air interfaces at the
% final time.

Y(end,1:ns+1)'
Y(end,ns*N+1:ns*N+ns+1)'
       
%% Plotting

% Plot time profiles at tissue-biofilm interface
 
figure(1)
subplot(4,2,1)
plot(T,Y(:,1))
xlabel('Time [h]')
ylabel('PA Biomass [g/L]')

subplot(4,2,2)
plot(T,Y(:,2))
xlabel('Time [h]')
ylabel('SA Biomass [g/L]')

subplot(4,2,3)
plot(T,Y(:,3))
xlabel('Time [h]')
ylabel('Glucose [mmol/L]')

subplot(4,2,4)
plot(T,Y(:,4))
xlabel('Time [h]')
ylabel('Oxygen [mmol/L]')

subplot(4,2,5)
plot(T,Y(:,5))
xlabel('Time [h]')
ylabel('Acetate [mmol/L]')

subplot(4,2,6)
plot(T,Y(:,6))
xlabel('Time [h]')
ylabel('Succinate [mmol/L]')

subplot(4,2,7)
plot(T,Y(:,7))
xlabel('Time [h]')
ylabel('Lactate [mmol/L]')

subplot(4,2,8)
plot(T,Y(:,8))
xlabel('Time [h]')
ylabel('Pyocyanin [mmol/L]')

% Plot time profiles in middle of biofilm

mz = round(N/2);

figure(2)
subplot(4,2,1)
plot(T,Y(:,ns*mz+1))
xlabel('Time [h]')
ylabel('PA Biomass [g/L]')

subplot(4,2,2)
plot(T,Y(:,ns*mz+2))
xlabel('Time [h]')
ylabel('SA Biomass [g/L]')

subplot(4,2,3)
plot(T,Y(:,ns*mz+3))
xlabel('Time [h]')
ylabel('Glucose [mmol/L]')

subplot(4,2,4)
plot(T,Y(:,ns*mz+4))
xlabel('Time [h]')
ylabel('Oxygen [mmol/L]')

subplot(4,2,5)
plot(T,Y(:,ns*mz+5))
xlabel('Time [h]')
ylabel('Acetate [mmol/L]')

subplot(4,2,6)
plot(T,Y(:,ns*mz+6))
xlabel('Time [h]')
ylabel('Succinate [mmol/L]')

subplot(4,2,7)
plot(T,Y(:,ns*mz+7))
xlabel('Time [h]')
ylabel('Lactate [mmol/L]')

subplot(4,2,8)
plot(T,Y(:,ns*mz+8))
xlabel('Time [h]')
ylabel('Pyocyanin [mmol/L]')

% Plot time profiles at biofilm-air interface

figure(3)
subplot(4,2,1)
plot(T,Y(:,ns*N+1))
xlabel('Time [h]')
ylabel('PA Biomass [g/L]')

subplot(4,2,2)
plot(T,Y(:,ns*N+2))
xlabel('Time [h]')
ylabel('SA Biomass [g/L]')

subplot(4,2,3)
plot(T,Y(:,ns*N+3))
xlabel('Time [h]')
ylabel('Glucose [mmol/L]')

subplot(4,2,4)
plot(T,Y(:,ns*N+4))
xlabel('Time [h]')
ylabel('Oxygen [mmol/L]')

subplot(4,2,5)
plot(T,Y(:,ns*N+5))
xlabel('Time [h]')
ylabel('Acetate [mmol/L]')

subplot(4,2,6)
plot(T,Y(:,ns*N+6))
xlabel('Time [h]')
ylabel('Succinate [mmol/L]')

subplot(4,2,7)
plot(T,Y(:,ns*N+7))
xlabel('Time [h]')
ylabel('Lactate [mmol/L]')

subplot(4,2,8)
plot(T,Y(:,ns*N+8))
xlabel('Time [h]')
ylabel('Pyocyanin [mmol/L]')

% Plot spatial profiles at final time

for i=1:N+1
    zloc(i) = (i-1)*zs;
    xbm(i) = Y(end,ns*(i-1)+1);
    zbm(i) = Y(end,ns*(i-1)+2);
    glc(i) = Y(end,ns*(i-1)+3);
    o2(i) = Y(end,ns*(i-1)+4);
    ace(i) = Y(end,ns*(i-1)+5);
    succ(i) = Y(end,ns*(i-1)+6);
    lac(i) = Y(end,ns*(i-1)+7);
    pyo(i) = Y(end,ns*(i-1)+8);
end

figure(4)
subplot(4,2,1)
plot(zloc,xbm)
xlabel('Location (micron)')
ylabel('PA biomass [g/L]')

subplot(4,2,2)
plot(zloc,zbm)
xlabel('Location (micron)')
ylabel('SA biomass [g/L]')

subplot(4,2,3)
plot(zloc,glc)
xlabel('Location (micron)')
ylabel('Glucose [mmol/L]')

subplot(4,2,4)
plot(zloc,o2)
xlabel('Location (micron)')
ylabel('Oxygen [mmol/L]')

subplot(4,2,5)
plot(zloc,ace)
xlabel('Location (micron)')
ylabel('Acetate [mmol/L]')

subplot(4,2,6)
plot(zloc,succ)
xlabel('Location (micron)')
ylabel('Succinate [mmol/L]')

subplot(4,2,7)
plot(zloc,lac)
xlabel('Location (micron)')
ylabel('Lactate [mmol/L]')

subplot(4,2,8)
plot(zloc,pyo)
xlabel('Location (micron)')
ylabel('Pyocyanin [mmol/L]')